dofile_once("data/scripts/lib/coroutines.lua")
dofile_once("data/scripts/lib/utilities.lua")

local delay_frames = 180
local entity_id = GetUpdatedEntityID()
local x, y = EntityGetTransform(entity_id)
SetRandomSeed(GameGetFrameNum() + x, GameGetFrameNum() + y)

-- probabilities are weighted
local events = {
  { -- Spawn a big black hole
    probability = 1,
    func = function()
      shoot_projectile(entity_id, "data/entities/projectiles/deck/black_hole_big.xml", x, y, 0, 0, false)
    end
  },
  { -- Spawn bombs
    probability = 1,
    func = function()
      for i=1,Random(1,3) do
        local random_angle = Randomf(-math.pi, math.pi)
        local throw_force = Random(30, 100)
        shoot_projectile(entity_id, "data/entities/projectiles/bomb.xml", x, y, math.cos(random_angle) * throw_force, math.sin(random_angle) * throw_force, false)
        wait(45)
      end
    end
  },
  { -- Shoot out 5 goldnuggets (50 gold)
    probability = 1,
    func = function()
      for i=1, 5 do
        local random_angle = Randomf(-math.pi, math.pi)
        local entity = EntityLoad("data/entities/items/pickup/goldnugget_10.xml", x, y)
        PhysicsApplyForce(entity, math.cos(random_angle) * 300, math.sin(random_angle) * 300)
        wait(20)
      end
    end
  },
  { -- Spawn tentacle portal
    probability = 1,
    func = function()
      shoot_projectile(entity_id, "data/entities/projectiles/deck/tentacle_portal.xml", x, y, 0, 0, false)
    end
  },
  { -- Fizzle, do nothing
    probability = 3,
    func = function()
      local poof = EntityLoad("data/entities/particles/poof_pink.xml", x, y)
      local particle_emitter_component = EntityGetFirstComponent(poof, "ParticleEmitterComponent")
      ComponentSetValue2(particle_emitter_component, "count_min", 10)
      ComponentSetValue2(particle_emitter_component, "count_max", 15)
      ComponentSetValue2(particle_emitter_component, "gravity", 0, 0)
      ComponentSetValue2(particle_emitter_component, "x_vel_min", -60)
      ComponentSetValue2(particle_emitter_component, "x_vel_max", 60)
      ComponentSetValue2(particle_emitter_component, "y_vel_min", -60)
      ComponentSetValue2(particle_emitter_component, "y_vel_max", 60)
    end
  },
  { -- Spawn random enemies
    probability = 1,
    func = function()
      local enemies = {
        { probability = 10, entity_file = "data/entities/animals/zombie.xml", max = 3 },
        { probability = 10, entity_file = "data/entities/animals/longleg.xml", max = 5 },
        { probability = 5, entity_file = "data/entities/animals/sniper.xml", max = 2 },
        { probability = 3, entity_file = "data/entities/animals/worm.xml", max = 1 },
        { probability = 1, entity_file = "data/entities/animals/worm_big.xml", max = 1 },
        { probability = 1, entity_file = "data/entities/animals/lukki/lukki.xml", max = 1 },
      }
      local rnd = random_create(GameGetFrameNum() + x, GameGetFrameNum() + y)
      local random_enemy = pick_random_from_table_weighted(rnd, enemies)
      for i=1, Random(1, random_enemy.max) do
        local random_angle = Randomf(-math.pi, math.pi)
        local entity = EntityLoad(random_enemy.entity_file, x, y)
        local character_data_component = EntityGetFirstComponentIncludingDisabled(entity, "CharacterDataComponent")
        if character_data_component then
          ComponentSetValue2(character_data_component, "mVelocity", math.cos(random_angle) * 300, math.sin(random_angle) * 300)
        end
        wait(45)
      end
    end
  },
  { -- Throw three random potions
    probability = 1,
    func = function()
      local throw_force = 125
      for i=1, 3 do
        local random_angle = Randomf(math.rad(60), math.rad(120)) + math.rad(180)
        local potion = EntityLoad("data/entities/items/pickup/potion.xml", x, y-i)
        PhysicsApplyForce(potion, math.cos(random_angle) * throw_force, math.sin(random_angle) * throw_force)
        wait(60)
      end
    end
  },
}

async(function()
  wait(delay_frames)
  local rnd = random_create(GameGetFrameNum() + x, GameGetFrameNum() + y)
  local random_event = pick_random_from_table_weighted(rnd, events)
  random_event.func()
  EntityKill(entity_id)
end)
