dofile_once("data/scripts/lib/utilities.lua")

local function create_explosion(x, y, power)
  if power < 1 then return end
  local new_entity = EntityCreateNew()
  EntitySetTransform(new_entity, x, y)
  local projectile_component = EntityAddComponent2(new_entity, "ProjectileComponent", {
    speed_min=160,
    speed_max=170,
    on_death_explode=true,
    on_death_gfx_leave_sprite=false,
    on_lifetime_out_explode=true,
    explosion_dont_damage_shooter=false,
    die_on_low_velocity=true,
    damage=0,
    on_collision_die=true,
    lifetime=1,
  })

  local config_explosion = {
    {
      damage=1,
      explosion_radius=28,
      explosion_sprite="data/particles/explosion_032.xml",
      load_this_entity="data/entities/particles/particle_explosion/main_gunpowder_medium.xml",
      audio_event_name="explosions/tnt",
      physics_explosion_power = { min = 1.2, max = 2.0 },
      max_durability_to_destroy=10,
    },
    {
      never_cache=true,
      damage=2,
      camera_shake=50,
      explosion_radius=60,
      explosion_sprite="data/particles/explosion_128.xml",
      explosion_sprite_lifetime=0,
      load_this_entity="data/entities/particles/particle_explosion/main_gunpowder_large.xml,data/entities/misc/loose_ground.xml",
      create_cell_probability=40,
      hole_destroy_liquid=false,
      hole_enabled=true,
      ray_energy=6000000,
      particle_effect=true,
      damage_mortals=true,
      physics_explosion_power = { min = 2.2, max = 3.6 },
      shake_vegetation=true,
      sparks_count_min=12,
      sparks_count_max=15,
      sparks_enabled=true,
      stains_enabled=true,
      stains_radius=15,
      max_durability_to_destroy=11,
    },
    {
      never_cache=true,
		  damage=8,
		  camera_shake=60,
		  explosion_radius=180,
		  explosion_sprite="data/particles/explosion_128.xml",
		  explosion_sprite_lifetime=0,
		  load_this_entity="data/entities/particles/particle_explosion/main_large_radius.xml,data/entities/misc/loose_ground.xml",
		  create_cell_probability=40,
		  hole_destroy_liquid=false,
		  hole_enabled=true,
		  ray_energy=6400000,
		  particle_effect=true,
      damage_mortals=true,
      physics_explosion_power = { min = 3.5, max = 7 },
		  shake_vegetation=true,
		  sparks_count_max=1500,
		  sparks_count_min=1600,
		  sparks_enabled=true,
		  stains_enabled=true,
		  stains_radius=35,
		  background_lightning_count=5,
		  max_durability_to_destroy=13,
    },
    {
      never_cache=true,
      damage=10,
      camera_shake=60,
      explosion_radius=250,
      explosion_sprite="data/particles/explosion_032.xml",
      explosion_sprite_lifetime=0,
	    load_this_entity="data/entities/particles/particle_explosion/main_large.xml,data/entities/misc/loose_chunks.xml",
      create_cell_probability=5,
      hole_destroy_liquid=false,
      hole_enabled=true,
      ray_energy=6700000,
      particle_effect=true,
      damage_mortals=true,
	    physics_explosion_power = { min = 4.5, max = 9 },
      shake_vegetation=true,
      sparks_count_max=1500,
      sparks_count_min=1600,
      sparks_enabled=true,
      stains_enabled=true,
      stains_radius=35,
      background_lightning_count=5,
	    max_durability_to_destroy=15,
      audio_event_name="explosions/nuke"
    }
  }

  config_explosion = config_explosion[math.min(#config_explosion, math.floor(power))]

  for k, v in pairs(config_explosion) do
    if type(v) == "table" then
      ComponentObjectSetValue2(projectile_component, "config_explosion", k, v.min, v.max)
    else
      ComponentObjectSetValue2(projectile_component, "config_explosion", k, v)
    end
  end
end

local entity_id = GetUpdatedEntityID()
local x, y = EntityGetTransform(entity_id)
local var_store_frame_start = get_variable_storage_component(entity_id, "MATRAN_TIMEBOMB_FRAME_START")
if var_store_frame_start then
  local frame_start = ComponentGetValue2(var_store_frame_start, "value_int")
  local seconds_active = (GameGetFrameNum() - frame_start) / 60
  seconds_active = math.max(0, seconds_active - 5)
  EntityRemoveComponent(owner, var_store_frame_start)
  if seconds_active > 0 then
    -- 0 - 5 seconds = 0
    -- 5 - 15 seconds = 1
    -- 15 - 25 seconds = 2
    -- 25 - 35 seconds = 3
    -- 35+ seconds = 4
    local explosion_power = seconds_active / 10 + 1
    create_explosion(x, y, explosion_power)
  end
  EntityRemoveComponent(entity_id, var_store_frame_start)
end
EntityRemoveComponent(entity_id, GetUpdatedComponentID())
