--[[
  In case you're reading this and wondering what the whole "if not EntityCreateNew then" is for,
  i'm running it externally using a luajit executable in a different environment, so I need to
  adapt the file paths for that. To determine wether it's running in noita or not,
  I'm checking if the EntityCreateNew function exists.
]]

print("============== Material Randomizer script start ==============")

function create_seed()
  local _,month,day,hour,minute,second = GameGetDateAndTimeUTC()
  return (month * 100000 + day * 10000 + hour * 1000 + minute * 100 + second) * 1000
end

local seed = 0
if EntityCreateNew then
  -- Check if we have a seed from an already started run
  seed = ModSettingGet("material_randomizer.current_seed")
  if not seed then
    if ModSettingGet("material_randomizer.fixed_seed") then
      local hex_string_seed = ModSettingGet("material_randomizer.seed")
      seed = tonumber(hex_string_seed, 16)
    else
      seed = create_seed()
    end
    ModSettingSet("material_randomizer.current_seed", seed)
  end
else
  seed = os.time()
end
-- seed = 1591729712
math.randomseed(seed)

-- Only used for developing using an external luajit.exe
local noita_path = "C:/Program Files (x86)/Steam/steamapps/common/Noita/"
local data_wak = noita_path .. "data/data.wak"
local output_file = "materials_new.xml"
local xml2lua
local handler
if not EntityCreateNew then
  xml2lua = dofile(noita_path .. "mods/material_randomizer/lib/xml2lua/xml2lua.lua")
  handler = dofile(noita_path .. "mods/material_randomizer/lib/xml2lua/xmlhandler/tree.lua")
else
  xml2lua = dofile("mods/material_randomizer/lib/xml2lua/xml2lua.lua")
  handler = dofile("mods/material_randomizer/lib/xml2lua/xmlhandler/tree.lua")
end
local xml
if not EntityCreateNew then
  local wak = dofile("wak.lua")
  local ar = wak.open(data_wak)
  local materials_file = ar:open("data/materials.xml")
  local materials_xml = materials_file:read()
  local parser = xml2lua.parser(handler)
  parser:parse(materials_xml)

  dofile("random_name_creator.lua")
  dofile("utilities.lua")
  function get_setting(setting_id)
    return ({
      randomize_properties_ground = true,
      randomize_properties_liquid = true,
      randomize_properties_sand = true,
      randomize_properties_gas = true,
      randomize_properties_fire = true,
      randomize_properties_solid = true,
      randomize_properties_breakable = true,

      randomize_name_ground = true,
      randomize_name_liquid = true,
      randomize_name_sand = true,
      randomize_name_gas = true,
      randomize_name_fire = true,
      randomize_name_solid = true,
      randomize_name_breakable = true,

      randomize_visual_ground = true,
      randomize_visual_liquid = true,
      randomize_visual_sand = true,
      randomize_visual_gas = true,
      randomize_visual_fire = true,
      randomize_visual_solid = true,
      randomize_visual_breakable = true,

      randomize_stain_liquid = true,
      randomize_stain_gas = true,

      randomize_ingestion_ground = true,
      randomize_ingestion_liquid = true,
      randomize_ingestion_sand = true,
      randomize_ingestion_fire = true,
      randomize_ingestion_solid = true,
      randomize_ingestion_breakable = true,

      exclude_material_blood = false,
      exclude_material_water = false,
      exclude_material_radioactive_liquid = false,

      ingestion_diarrhea = true,
      ingestion_bleed_gold = true,
      ingestion_minor_regeneration = true,
      ingestion_explosive_farts = true,
      ingestion_heavy = true,
      ingestion_incontinence = true,
      ingestion_powerful_kicks = true,
      ingestion_timebomb = true,
      ingestion_low_gravity = true,
      ingestion_gain_max_hp = true,
      ingestion_halve_mana_cost = true,
      ingestion_long_arm = true,
      ingestion_anomaly_catalyst = true,
      ingestion_noodle_arm = true,
      ingestion_evaporate_liquids = true,
      -- Vanilla now
      ingestion_drunk = true,
      ingestion_trip = true,
      ingestion_nightvision = true,
      ingestion_food_poisoning = true,
      ingestion_movement_slower = true,
      ingestion_freezing = true,
      ingestion_farts = true,
      ingestion_on_fire = true,

      stains_lose_levitation = true,
      stains_infinite_levitation = true,
      stains_restore_spell_uses = true,
      stains_rapidfire = true,
      stains_impotence = true,
      stains_blindness = true,
      stains_attract_projectiles = true,
      stains_slippery_hands = true,
      stains_goldrush = true,
      stains_projectile_repulsion = true,
      stains_defense_up = true,
      stains_attractive = true,
      stains_floaty = true,
      stains_radioactive_matran = true,
      stains_blackhole = true,
      stains_demolition = true,
      -- Vanilla now
      stains_wet = true,
      stains_oiled = true,
      stains_bloody = true,
      stains_slimy = true,
      stains_radioactive = true,
      stains_alcoholic = true,
      stains_poisoned = true,
      stains_teleportation = true,
      stains_hp_regeneration = true,
      stains_polymorph = true,
      stains_polymorph_random = true,
      stains_berserk = true,
      stains_charm = true,
      stains_invisibility = true,
      stains_confusion = true,
      stains_movement_faster_2x = true,
      stains_worm_attractor = true,
      stains_protection_all = true,
      stains_mana_regeneration = true,
    })[setting_id]
  end
else
  dofile_once("mods/material_randomizer/files/random_name_creator.lua")
  dofile_once("mods/material_randomizer/files/utilities.lua")
  xml = ModTextFileGetContent("data/materials.xml")
  local parser = xml2lua.parser(handler)
  parser:parse(xml)
end

function shuffle(tbl)
  for i = #tbl, 2, -1 do
    local j = math.random(i)
    tbl[i], tbl[j] = tbl[j], tbl[i]
  end
  return tbl
end

function should_shuffle(name, tags, attr)
  if attr.vegetation_sprite ~= nil then
    return false
  end
  --exclusion: potion_glass_box2d maybe?
  if (not tags.box2d) or (tags.box2d and tags.static) then
    return true
  end
  return false
end

function parse_tags(tags)
  local output = {}
  if tags == nil then
    return output
  end
  local a = string.gmatch(tags, "%[([%w_]+)%]")
  for i in a do
    output[i] = true
  end
  return output
end

function get_child_materials(materials, material)
  local children = {}
  for name, parent in pairs(materials) do
    if parent == material then
      table.insert(children, name)
    end
  end
  return children
end

-- Returns parent materials as an array, for instance:
-- input: pea_soup
-- output: { "slime", "slime_green", "pea_soup" }
function get_parent_chain(materials, material)
  local chain = {}
  local orig_mat = material
  while materials[material] ~= "none" do
    table.insert(chain, materials[material])
    material = materials[material]
  end
  -- reverse the table
  for i=1, math.floor(#chain / 2) do
    local tmp = chain[i]
    chain[i] = chain[#chain - i + 1]
    chain[#chain - i + 1] = tmp
  end
  table.insert(chain, orig_mat)
  return chain
end

function build_inheritance_tree(materials)
  local tree = {}
  -- First, collect all original materials
  for name, parent in pairs(materials) do
    local parent_chain = get_parent_chain(materials, name)
    local current = tree
    for level, mat in ipairs(parent_chain) do
      -- We start of at the original material and walk our way downward towards this child
      -- If the child was parsed first, and the uppermost parent doesn't exist yet, create it
      if current[mat] == nil then
        current[mat] = {
          _name = mat,
          _parentElement = current,
          _rootElement = tree[parent_chain[1]],
          _parent = materials[mat], -- _parent will be "none" when level == 1
        }
      end
      current = current[mat]
    end
  end
  return tree
end

-- Gather all material names and parents
function gather_parents()
  local parents = {}
  local names_already_encountered = {}
  for i, p in pairs(handler.root.Materials) do
    if i == "CellDataChild" or i == "CellData" then
      for i2, p2 in pairs(handler.root.Materials[i]) do
        local name = p2._attr.name
        local _parent = p2._attr._parent
        -- Protect against invalid double named materials
        if not names_already_encountered[name] then
          parents[name] = _parent or "none"
          names_already_encountered[name] = true
        end
      end
    end
  end
  return parents
end

local parents = gather_parents()
local tree = build_inheritance_tree(parents)

--[[

tree = {
  wood = {
    _parent = "none",
    _name = "wood",
    _XMLElement = ???,
    wood_wall = {
      _parent = "wood",
      _name = "wood_wall",
      _XMLElement = ???,
    },
    templebrick_box2d = {
      _parent = "wood",
      _name = "templebrick_box2d",
      _XMLElement = ???,
    },
  }
}

]]

function get_tree_node(tree, materials, material_name)
  local chain = get_parent_chain(materials, material_name)
  local node = tree
  for i, v in ipairs(chain) do
    node = node[v]
  end
  return node
end

-- Add xml node references to tree
for i, p in pairs(handler.root.Materials) do
  if i == "CellDataChild" or i == "CellData" then
    for i2, p2 in pairs(handler.root.Materials[i]) do
      local name = p2._attr.name
      local node = get_tree_node(tree, parents, name)
      node._XMLElement = p2
    end
  end
end

function clone_table(t)
  local out = {}
  for k, v in pairs(t) do
    if type(v) == "table" then
      out[k] = clone_table(v)
    else
      out[k] = v
    end
  end
  return out
end

-- merge t2 into t1, values t2 will overwrite values in t1
function inherit_element(t1, t2)
  for k, v in pairs(t2) do
    if t1[k] == nil then
      if type(v) == "table" then
        t1[k] = clone_table(v)
      else
        t1[k] = v
      end
    else
      if type(v) == "table" and k == "_attr" then
        inherit_element(t1[k], v)
      elseif k == "tags" then
        t1[k] = t1[k] .. "," .. v
      end
    end
  end
end
-- Returns a new object with the same structure as an XML element, that has
-- all inherited attributes etc
function get_inherited_element(node)
  local out = clone_table(node._XMLElement)
  local element = node._parentElement
  while element and element._XMLElement do
    inherit_element(out, element._XMLElement)
    element = element._parentElement
  end
  -- require"pl/pretty".dump(out, "shitxxy.lua")
  return out
end

function material_get_type(element)
  local a = element._attr
  if a.cell_type == "liquid" and (a.liquid_sand == nil or a.liquid_sand == "0") then
    return "liquid"
  elseif a.cell_type == "liquid" and a.liquid_sand == "1"
         and (a.liquid_static == nil or a.liquid_static == "0")
         and (a.liquid_sticks_to_ceiling == nil or a.liquid_sticks_to_ceiling == "0") then
    return "sand"
  elseif a.cell_type == "solid" then
    -- Used for physics stuff and pixel scenes etc
    return "solid"
  elseif (a.cell_type == "solid" and a.solid_break_to_type ~= nil) or a.convert_to_box2d_material then
    return "breakable"
  elseif a.cell_type == "liquid" and a.liquid_sand == "1" then
    -- Used by the world by wang generator and biomes
    return "ground"
  elseif a.cell_type == "fire" then
    return "fire"
  elseif a.cell_type == "gas" then
    return "gas"
  end
  return "unknown"
end


local random_textures = {}
local custom_texture_types = {
  liquid = {
    "endslime","endslime_translucent","honey","honey_translucent","midas_precursor","midas_precursor_translucent", "slime",
    "slime_translucent","waterstone","waterstone_translucent","wormmeat","wormmeat_translucent","yellowblood","yellowblood_translucent"
  },
  sand = {
    "sand_surface", "root_growth", "meteor", "lavarock", "gunpowder", "gold_dull", "gold_bright", "dirt"
  },
}


function table.contains(t, element)
  for i, v in ipairs(t) do
    if v == element then
      return true
    end
  end
  return false
end

local categories = {} -- contains references to all materials based on their category (liquid, sand, etc)
local exclude_materials = { "potion_glass_box2d", "gold_box2d", "bloodgold_box2d", "gem_box2d", "fuse", "urine", "poo" }
if get_setting("exclude_material_blood") then
  table.insert(exclude_materials, "blood")
end
if get_setting("exclude_material_water") then
  table.insert(exclude_materials, "water")
end
if get_setting("exclude_material_radioactive_liquid") then
  table.insert(exclude_materials, "radioactive_liquid")
end

-- First loop, collect all materials and textures
local names_already_encountered = {}
for node_name, nodes in pairs(handler.root.Materials) do
  if node_name == "CellDataChild" or node_name == "CellData" then
    for i, element in pairs(nodes) do
      local name = element._attr.name
      if not names_already_encountered[name] then
        names_already_encountered[name] = true
        if not table.contains(exclude_materials, name) then
          local inherited_element = get_inherited_element(get_tree_node(tree, parents, name))
          local category = material_get_type(inherited_element)
          categories[category] = categories[category] or {}
          random_textures[category] = random_textures[category] or {}

          table.insert(categories[category], element)
          if inherited_element.Graphics and inherited_element.Graphics._attr.texture_file then
            if inherited_element.Graphics._attr.texture_file == "" then
              element.Graphics._attr.texture_file = nil
            elseif inherited_element.Graphics.Edge then
              table.insert(random_textures[category], {
                graphics = inherited_element.Graphics,
              })
            else
              table.insert(random_textures[category], inherited_element.Graphics._attr.texture_file)
            end
          end
        end
      else
        -- Material exists twice, remove the second one since it won't have any effect and
        -- just causes wang color collisions
        table.remove(nodes, i)
      end
    end
  end
end

for category, names in pairs(custom_texture_types) do
  random_textures[category] = {}
  for i, name in ipairs(names) do
    for i=1,10 do
      table.insert(random_textures[category], "mods/material_randomizer/files/materials_gfx/" .. name .. "/" .. i .. ".png")
    end
  end
  shuffle(random_textures[category])
end

local audio_events = {
  solid = { "rock", "wood", "ice", "sand", "slime", "meat", "gravel", "organicbouncy", "metalhollow_gold",
            "bone", "metalhollow", "gold", "glass", "metalhollow_barrel", "snow", "organicsoft" },
  event = { "rock", "sand", "gravel", "slime", "metalhollow_gold", "metalhollow", "snow" },
  wall = { "gravel", "rock", "woodwall", "ice", "sand", "slime", "meat", "organicbouncy", "metalwall", "bone",
           "glass", "metalhollow", "metalhollow_barrel", "snow", "organicsoft" }
}

function add_tag(element, tag)
  if not element._attr.tags then
    element._attr.tags = string.format("[%s]", tag)
  else
    local existing_tags = parse_tags(element._attr.tags)
    for tag2, _ in pairs(existing_tags) do
      if tag == tag2 then
        return
      end
    end
    element._attr.tags = element._attr.tags .. string.format(",[%s]", tag)
  end
end

function remove_tag(element, tag)
  if element._attr.tags then
    element._attr.tags = string.gsub(element._attr.tags, "%s*,?%s*%["..tag.."%]", "")
    element._attr.tags = string.gsub(element._attr.tags, "^%s*,", "")
  end
end

local randomizers = {
  liquid = function(element, inherited_element)
    add_tag(element, "liquid")
    element._attr.liquid_stains = "1"
    element._attr.on_fire = "0" -- Makes it be able to put out burning entities
    if get_setting("randomize_visual_liquid") then
      element.ParticleEffect = nil
      if math.random(100) <= 20 then -- 20% chance to give it a particle effect
        local sideways_vel = randf(10)
        element.ParticleEffect = {
          _attr = {
            ["vel.y"] = "-13.14",
            ["vel_random.min_x"] = -sideways_vel,
            ["vel_random.max_x"] = sideways_vel,
            ["vel_random.min_y"] = randf(-20, -2.8),
            ["vel_random.max_y"] = -2.8,
            ["lifetime.min"] = randf(0.2, 0.7),
            ["lifetime.max"] = randf(1.0, 2.0),
            ["gravity.y"] = randf(-20, 80),
            ["render_on_grid"] = "1",
            ["airflow_force"] = randf(1),
            ["airflow_scale"] = randf(1),
            ["friction"] = randf(0.3),
            ["probability"] = randf(0.01),
            ["count.min"] = "0",
            ["count.max"] = "1",
          }
        }
      end
      -- Give it a random color or texture
      local random_texture
      if rand(100) <= 5 then
        random_texture = table.remove(random_textures.liquid, rand(1, #random_textures.liquid))
      end

      element.Graphics = { _attr = {} }
      if random_texture then
        element.Graphics._attr.texture_file = random_texture
      end
      element.Graphics._attr.color = argb2hex(rand(100, 210), hsv2rgb(rand(360), randf(0.6, 1.0), randf(0.6, 1.0)))
      element._attr.gfx_glow_color = nil
      element._attr.gfx_glow = nil
      if rand(100) <= 20 then
        element._attr.gfx_glow = rand(0, 255)
      end
      -- if rand(100) <= 100 then
      --   element._attr.liquid_stains = "2"
      -- end
    end

    -- Make it flammable
    if get_setting("flammable_liquids") then
      if rand(100) <= 5 then
        element._attr.burnable = "1"
        element._attr.requires_oxygen = rand(0, 1)
        if rand(100) <= 50 then
          element._attr.generates_flames = rand(1, 20)
        else
          element._attr.generates_smoke = rand(1, 20)
        end
        if rand(100) <= 20 then
          element._attr.autoignition_temperature = rand(1, 20)
        else
          element._attr.autoignition_temperature = rand(20, 99)
        end
      end
      element._attr.fire_hp = rand(100, 500)
    end

    if get_setting("randomize_properties_liquid") then
      if rand(100) <= 10 then
        element._attr.liquid_slime = "1"
      end
      -- if rand(100) <= 2 then
      --   element._attr.always_ignites_damagemodel = "1"
      -- end
      element._attr.density = rand(1, 5)
      if get_setting("change_gravity") then
        element._attr.liquid_gravity = randf(0.5, 10.0)
      end
      element._attr.liquid_sprite_stain_ignited_drop_chance = rand(0, 10)
      -- element._attr.liquid_sprite_stains_status_threshold = randf(0.1, 0.5)
      element._attr.liquid_sprite_stain_shaken_drop_chance = rand(1, rand(1, 5)) -- Reduce the chance to get higher values

      if rand(100) <= 1 then
        add_tag(element, "densify")
      end
      if rand(100) <= 1 then
        add_tag(element, "matran_sandify")
      end
      element._attr.danger_radioactive = nil
      element._attr.danger_fire = nil
      element._attr.danger_water = nil
      element._attr.danger_poison = nil
      remove_tag(element, "lava")
      remove_tag(element, "acid")
    end
    -- element._attr.ui_name = "Liquid?" .. inherited_element._attr.ui_name
  end,
  sand = function(element, inherited_element)
    if get_setting("randomize_visual_sand") then
      local random_texture = #random_textures.sand > 0 and table.remove(random_textures.sand, rand(1, #random_textures.sand)) or nil
      element.Graphics = { _attr = { texture_file = random_texture } }
    end
    if get_setting("randomize_properties_sand") then
      element._attr.density = rand(3, 8)

      local durability_roll = rand(100)
      if durability_roll <= 15 then -- 15% chance
        element._attr.durability = rand(1, 7)
      elseif durability_roll <= 80 then -- 65% chance
        element._attr.durability = rand(7, 12)
      else -- 20% chance
        element._attr.durability = rand(7, 20)
      end

      if get_setting("change_gravity") then
        element._attr.liquid_gravity = randf(0.5, 20.0)
      end
      -- if rand(100) <= 2 then
      --   element._attr.always_ignites_damagemodel = "1"
      -- end
      element._attr.requires_oxygen = rand(0, 1)
      element._attr.temperature_of_fire = rand(5, 200)
      element._attr.autoignition_temperature = rand(10, 99)
      if rand(100) <= 20 then
        element._attr.solid_friction = randf(0.05, 1.0)
      end
      element._attr.electrical_conductivity = rand(0, 1)
      element._attr.liquid_flow_speed = randf(0.0, 1.0)

      element._attr.danger_radioactive = nil
      element._attr.danger_fire = nil
      element._attr.danger_water = nil
      element._attr.danger_poison = nil
      -- element._attr.ui_name = "Sand?" .. inherited_element._attr.ui_name
    end
  end,
  solid = function(element, inherited_element)
    if get_setting("randomize_visual_solid") then
      local random_texture = #random_textures.solid > 0 and table.remove(random_textures.solid, rand(1, #random_textures.solid)) or nil
      if type(random_texture) == "table" then
        element.Graphics = random_texture.graphics
      else
        element.Graphics = { _attr = { texture_file = random_texture } }
      end
    end
    if get_setting("randomize_properties_solid") then
      element._attr.density = rand(5, 12)

      local durability_roll = rand(100)
      if durability_roll <= 15 then -- 15% chance
        element._attr.durability = rand(1, 7)
      elseif durability_roll <= 80 then -- 65% chance
        element._attr.durability = rand(7, 12)
      else -- 20% chance
        element._attr.durability = rand(7, 20)
      end

      element._attr.solid_restitution = randf(0.0, 0.8)
      element._attr.solid_friction = randf(0.05, 1.0)
      element._attr.solid_go_through_sand = rand(0, 1)
      if get_setting("change_gravity") then
        element._attr.solid_gravity_scale = randf(0, 1.3)
      end
      element._attr.hp = rand(100, 5000)

      if rand(100) <= 10 then
        element._attr.burnable = "1"
        element._attr.requires_oxygen = rand(0, 1)
        if rand(100) <= 50 then
          element._attr.generates_flames = rand(1, 20)
        else
          element._attr.generates_smoke = rand(1, 20)
        end
        if rand(100) <= 20 then
          element._attr.autoignition_temperature = rand(1, 20)
        else
          element._attr.autoignition_temperature = rand(20, 99)
        end
      end
      -- if rand(100) <= 2 then
      --   element._attr.always_ignites_damagemodel = "1"
      -- end
      element._attr.fire_hp = rand(100, 500)
      local tags = parse_tags(inherited_element._attr.tags)
      if not tags.box2d and rand(100) <= 100 then
        element._attr.solid_on_break_explode = "1"
        local explosion_power = rand(5, 100)
        element.ExplosionConfig = { _attr = {
          camera_shake = math.floor(explosion_power / 3),
          cell_explosion_power = explosion_power,
          cell_explosion_damage_required = rand(5, 50),
          cell_explosion_radius_min = rand(2, 10),
          ray_energy = "50000",
        } }
      end
      element._attr.danger_radioactive = nil
      element._attr.danger_fire = nil
      element._attr.danger_water = nil
      element._attr.danger_poison = nil

      element._attr.audio_physics_material_solid = audio_events.solid[rand(1, #audio_events.solid)]
      element._attr.audio_physics_material_event = audio_events.event[rand(1, #audio_events.event)]
      element._attr.audio_physics_material_wall = audio_events.wall[rand(1, #audio_events.wall)]
    end
  end,
  ground = function(element, inherited_element)
    if get_setting("randomize_visual_ground") then
      local random_texture = #random_textures.ground > 0 and table.remove(random_textures.ground, rand(1, #random_textures.ground)) or nil
      if type(random_texture) == "table" then
        element.Graphics = random_texture.graphics
      else
        element.Graphics = { _attr = { texture_file = random_texture } }
      end
    end

    if get_setting("randomize_properties_ground") then
      local durability_roll = rand(100)
      if durability_roll <= 15 then -- 15% chance
        element._attr.durability = rand(1, 7)
      elseif durability_roll <= 80 then -- 65% chance
        element._attr.durability = rand(7, 12)
      else -- 20% chance
        element._attr.durability = rand(7, 20)
      end

      if rand(100) <= 2 then
        element._attr.burnable = 1
        element._attr.requires_oxygen = rand(100) <= 90 and 1 or 0
        element._attr.autoignition_temperature = rand(5, 80)
        element._attr.generates_smoke = rand(1, 5)
        element._attr.temperature_of_fire = rand(10, 200)
        if rand(100) <= 50 then
          -- Slow burning fire
          element._attr.fire_hp = rand(500, 2000)
          element._attr.generates_flames = rand(5, 15)
        else
          -- Fast burning fire
          element._attr.fire_hp = rand(1, 100)
          element._attr.generates_flames = rand(1, 5)
        end
      end

      -- if rand(100) <= 2 then
      --   element._attr.always_ignites_damagemodel = "1"
      -- end
      element._attr.density = rand(5, 12)
      if get_setting("change_gravity") then
        element._attr.liquid_gravity = randf(0.5, 20.0)
      end
      if rand(100) <= 10 then
        element._attr.electrical_conductivity = "1"
      end
      if rand(100) <= 10 then
        if not element._attr.tags then
          element._attr.tags = "[shrinking]"
        else
          element._attr.tags = element._attr.tags .. ",[shrinking]"
        end
      end
    end
  end,
  breakable = function(element, inherited_element)
    if get_setting("randomize_visual_breakable") then
      local random_texture = #random_textures.breakable > 0 and table.remove(random_textures.breakable, rand(1, #random_textures.breakable)) or nil
      if type(random_texture) == "table" then
        element.Graphics = random_texture.graphics
      elseif random_texture then
        element.Graphics = { _attr = { texture_file = random_texture } }
      else
        -- This shouldn't really happen?
        element.Graphics = { _attr = { color = argb2hex(rand(100, 255), hsv2rgb(rand(360), randf(0.6, 1.0), randf(0.6, 1.0))) } }
      end
      element._attr.gfx_glow = rand(0, 100)
    end
    if get_setting("randomize_properties_breakable") then
      -- Turn it into static liquid
      if rand(100) <= 10 then
        element._attr = {
          _parent = element._attr._parent,
          name = element._attr.name,
          ui_name = element._attr.ui_name,
          wang_color = element._attr.wang_color,
          cell_type = "liquid",
          liquid_static = "1",
          liquid_slime = "1",
          liquid_stains = "1",
          density = rand(3, 10),
        }
        element.StatusEffects = { Stains = { StatusEffect = { _attr = { type = "SLIMY" } } } }
      end
      element._attr.electrical_conductivity = rand(100) <= 10 and 1 or 0

      local break_type_roll = rand(100)
      if break_type_roll <= 10 then
        -- Breaks to solid material
        element._attr.convert_to_box2d_material = random_from_table(categories.solid)._attr.name
      elseif break_type_roll <= 20 then
        element._attr.solid_on_break_explode = 1
        element._attr.solid_on_collision_splash_power = 1
        local explosion_power = rand(5, 30)
        element.ExplosionConfig = { _attr = {
          camera_shake = math.floor(explosion_power / 3),
          cell_explosion_power = explosion_power,
          cell_explosion_damage_required = rand(5, 50),
          cell_explosion_radius_min = rand(2, 10),
          ray_energy = "50000",
        } }
      end

      element._attr.crackability = rand(10, 100)
      element._attr.slippery = rand(0, 1)

      element._attr.burnable = rand(0, 1)
      element._attr.temperature_of_fire = rand(20, 200)
      element._attr.autoignition_temperature = rand(10, 80)

      element._attr.audio_physics_material_solid = audio_events.solid[rand(1, #audio_events.solid)]
      element._attr.audio_physics_material_event = audio_events.event[rand(1, #audio_events.event)]
      element._attr.audio_physics_material_wall = audio_events.wall[rand(1, #audio_events.wall)]

      element._attr.danger_radioactive = nil
      element._attr.danger_fire = nil
      element._attr.danger_water = nil
      element._attr.danger_poison = nil
      -- element._attr.ui_name = "Breakable?" .. inherited_element._attr.ui_name
    end
  end,
  fire = function(element, inherited_element)
    if get_setting("randomize_visual_fire") then
      element._attr.gfx_glow = rand(0, 255)
      element.Graphics = { _attr = {
        color = argb2hex(rand(100, 255), hsv2rgb(rand(360), randf(0.5, 1.0), randf(0.8, 1.0)))
      } }
    end
    if get_setting("randomize_properties_fire") then
      element._attr.temperature_of_fire = rand(5, 200)
    end
  end,
  gas = function(element, inherited_element)
    if get_setting("randomize_visual_gas") then
      local sideways_vel = randf(15)
      element.ParticleEffect = {
        _attr = {
          ["vel.y"] = randf(-5, -15),
          ["vel_random.min_x"] = -sideways_vel,
          ["vel_random.max_x"] = sideways_vel,
          ["vel_random.min_y"] = randf(-20, -2.8),
          ["vel_random.max_y"] = -2.7,
          ["lifetime.min"] = randf(0.2, 0.7),
          ["lifetime.max"] = randf(1.0, 2.0),
          ["gravity.y"] = randf(0, -60),
          ["render_on_grid"] = rand(0, 1),
          ["airflow_force"] = randf(1),
          ["airflow_scale"] = randf(1),
          ["friction"] = randf(0.3),
          ["probability"] = randf(0.05),
          ["count.min"] = rand(0, 1),
          ["count.max"] = "1",
        }
      }
      element.Graphics = { _attr = {
        color = argb2hex(rand(130, 230), hsv2rgb(rand(360), randf(0.5, 1.0), randf(0.8, 1.0)))
      } }
    end
    if get_setting("randomize_properties_gas") then
      if rand(100) <= 1 then
        added_matran_corrosive_gas = true
        add_tag(element, "matran_corrosive_gas")
      end
      element._attr.lifetime = rand(50, 300)
      element._attr.fire_hp = rand(50, 1000)
      element._attr.requires_oxygen = rand(0, 1)
      element._attr.temperature_of_fire = rand(50, 200)
      element._attr.autoignition_temperature = rand(1, 50)
      -- Can be static and even solid (platform type 1)
    end
  end,
  unknown = function(element, inherited_element)
  end,
}

function randomize_material(element)
  local inherited_element = get_inherited_element(get_tree_node(tree, parents, element._attr.name))
  local material_type = material_get_type(inherited_element)
  randomizers[material_type](element, inherited_element)
end

local random_material_names = create_random_material_names(1000)

-- Second loop, randomize all materials
for node_name, nodes in pairs(handler.root.Materials) do
  if node_name == "CellDataChild" or node_name == "CellData" then
    for _, element in pairs(nodes) do
      if not table.contains(exclude_materials, element._attr.name) then
        local inherited_element = get_inherited_element(get_tree_node(tree, parents, element._attr.name))
        local material_type = material_get_type(inherited_element)
        randomizers[material_type](element, inherited_element)
        element._attr.always_ignites_damagemodel = "0"
        if get_setting("randomize_name_" .. material_type) then
          element._attr.ui_name = table.remove(random_material_names, rand(1, #random_material_names))
        end
      end
    end
  end
end

-- Writes shimple inheritance chain to file
function print_tree(tree, level, f)
  local spaces = string.rep(" ", level*2)
  for k, v in pairs(tree) do
    if string.sub(k,1,1) ~= "_" then
      f:write(spaces .. k .. "\n")
      print_tree(v, level + 1, f)
    end
  end
end

-- local f = assert(io.open("print_tree.txt", "w"))
-- print_tree(tree, 0, f)
-- f:close()

local ingestion_effects = {}
local vanilla_ingestion_effects = {
  { name = "INGESTION_DRUNK", amount_min = 0.2, amount_max = 0.4, setting_id = "ingestion_drunk" },
  { name = "TRIP", amount_min = 0.2, amount_max = 0.4, setting_id = "ingestion_trip" },
  { name = "NIGHTVISION", amount_min = 0.4, amount_max = 1.0, setting_id = "ingestion_nightvision" },
  { name = "FOOD_POISONING", amount_min = 0.3, amount_max = 0.5, setting_id = "ingestion_food_poisoning" },
  { name = "INGESTION_MOVEMENT_SLOWER", amount_min = 0.1, amount_max = 0.3, setting_id = "ingestion_movement_slower" },
  -- { name = "INGESTION_DAMAGE", amount_min = 0.3, amount_max = 0.6, setting_id = "ingestion_ingestion_damage" },
  -- { name = "INGESTION_EXPLODING", amount_min = 0.2, amount_max = 0.5, setting_id = "ingestion_ingestion_exploding" },
  { name = "INGESTION_FREEZING", amount_min = 0.1, amount_max = 0.3, setting_id = "ingestion_freezing" },
  { name = "FARTS", amount_min = 0.3, amount_max = 0.7, setting_id = "ingestion_farts" },
  { name = "INGESTION_ON_FIRE", amount_min = 0.1, amount_max = 0.2, setting_id = "ingestion_on_fire" },
}

local custom_ingestion_effects = {
  { name = "MATRAN_DIARRHEA", amount_min = 0.4, amount_max = 0.7, setting_id = "ingestion_diarrhea" },
  { name = "MATRAN_BLEED_GOLD", amount_min = 0.5, amount_max = 1.0, setting_id = "ingestion_bleed_gold" },
  { name = "MATRAN_MINOR_REGENERATION", amount_min = 0.2, amount_max = 0.6, setting_id = "ingestion_minor_regeneration" },
  { name = "MATRAN_EXPLOSIVE_FARTS", amount_min = 0.4, amount_max = 1.0, setting_id = "ingestion_explosive_farts" },
  { name = "MATRAN_HEAVY", amount_min = 0.3, amount_max = 0.5, setting_id = "ingestion_heavy" },
  { name = "MATRAN_INCONTINENCE", amount_min = 0.2, amount_max = 0.7, setting_id = "ingestion_incontinence" },
  { name = "MATRAN_POWERFUL_KICKS", amount_min = 0.4, amount_max = 0.8, setting_id = "ingestion_powerful_kicks" },
  { name = "MATRAN_TIMEBOMB", amount_min = 0.2, amount_max = 0.4, setting_id = "ingestion_timebomb" },
  { name = "MATRAN_LOW_GRAVITY", amount_min = 0.4, amount_max = 0.8, setting_id = "ingestion_low_gravity" },
  { name = "MATRAN_GAIN_MAX_HP", amount_min = 0.1, amount_max = 0.3, setting_id = "ingestion_gain_max_hp" },
  { name = "MATRAN_HALVE_MANA_COST", amount_min = 0.3, amount_max = 0.6, setting_id = "ingestion_halve_mana_cost" },
  { name = "MATRAN_LONG_ARM", amount_min = 0.2, amount_max = 0.5, setting_id = "ingestion_long_arm" },
  { name = "MATRAN_ANOMALY_CATALYST", amount_min = 0.2, amount_max = 0.8, setting_id = "ingestion_anomaly_catalyst" },
  { name = "MATRAN_NOODLE_ARM", amount_min = 0.2, amount_max = 0.5, setting_id = "ingestion_noodle_arm" },
  { name = "MATRAN_EVAPORATE_LIQUIDS", amount_min = 0.2, amount_max = 0.8, setting_id = "ingestion_evaporate_liquids" },
}

for i, effect in ipairs(vanilla_ingestion_effects) do
  if get_setting(effect.setting_id) then
    table.insert(ingestion_effects, effect)
  end
end
for i, effect in ipairs(custom_ingestion_effects) do
  if get_setting(effect.setting_id) then
    table.insert(ingestion_effects, effect)
  end
end

local stains_effects = {}
local vanilla_stains_effects = {
  { name = "WET", drop_chance_min = 1, drop_chance_max = 5, setting_id = "stains_wet" },
  { name = "OILED", drop_chance_min = 1, drop_chance_max = 5, setting_id = "stains_oiled" },
  { name = "BLOODY", drop_chance_min = 1, drop_chance_max = 5, setting_id = "stains_bloody" },
  { name = "SLIMY", drop_chance_min = 1, drop_chance_max = 5, setting_id = "stains_slimy" },
  { name = "RADIOACTIVE", drop_chance_min = 1, drop_chance_max = 5, setting_id = "stains_radioactive" },
  { name = "ALCOHOLIC", drop_chance_min = 1, drop_chance_max = 5, setting_id = "stains_alcoholic" },
  { name = "POISONED", drop_chance_min = 3, drop_chance_max = 5, setting_id = "stains_poisoned" },
  { name = "TELEPORTATION", drop_chance_min = 1, drop_chance_max = 5, setting_id = "stains_teleportation" },
  { name = "HP_REGENERATION", drop_chance_min = 4, drop_chance_max = 5, setting_id = "stains_hp_regeneration" },
  { name = "POLYMORPH", drop_chance_min = 4, drop_chance_max = 5, setting_id = "stains_polymorph" },
  { name = "POLYMORPH_RANDOM", drop_chance_min = 4, drop_chance_max = 5, setting_id = "stains_polymorph_random" },
  { name = "BERSERK", drop_chance_min = 1, drop_chance_max = 5, setting_id = "stains_berserk" },
  { name = "CHARM", drop_chance_min = 1, drop_chance_max = 5, setting_id = "stains_charm" },
  { name = "INVISIBILITY", drop_chance_min = 1, drop_chance_max = 5, setting_id = "stains_invisibility" },
  -- { name = "ON_FIRE", drop_chance_min = 1, drop_chance_max = 5, setting_id = "stains_on_fire" },
  { name = "CONFUSION", drop_chance_min = 3, drop_chance_max = 5, setting_id = "stains_confusion" },
  { name = "MOVEMENT_FASTER_2X", drop_chance_min = 1, drop_chance_max = 5, setting_id = "stains_movement_faster_2x" },
  { name = "WORM_ATTRACTOR", drop_chance_min = 1, drop_chance_max = 5, setting_id = "stains_worm_attractor" },
  { name = "PROTECTION_ALL", drop_chance_min = 3, drop_chance_max = 5, setting_id = "stains_protection_all" },
  { name = "MANA_REGENERATION", drop_chance_min = 1, drop_chance_max = 5, setting_id = "stains_mana_regeneration" },
}

local custom_stains_effects = {
  { name = "MATRAN_LOSE_LEVITATION", drop_chance_min = 1, drop_chance_max = 5, setting_id = "stains_lose_levitation" },
  { name = "MATRAN_INFINITE_LEVITATION", drop_chance_min = 1, drop_chance_max = 5, setting_id = "stains_infinite_levitation" },
  { name = "MATRAN_RESTORE_SPELL_USES", drop_chance_min = 5, drop_chance_max = 10, setting_id = "stains_restore_spell_uses" },
  { name = "MATRAN_RAPIDFIRE", drop_chance_min = 3, drop_chance_max = 5, setting_id = "stains_rapidfire" },
  { name = "MATRAN_IMPOTENCE", drop_chance_min = 5, drop_chance_max = 7, setting_id = "stains_impotence" },
  { name = "MATRAN_BLINDNESS", drop_chance_min = 5, drop_chance_max = 7, setting_id = "stains_blindness" },
  { name = "MATRAN_ATTRACT_PROJECTILES", drop_chance_min = 1, drop_chance_max = 5, setting_id = "stains_attract_projectiles" },
  { name = "MATRAN_SLIPPERY_HANDS", drop_chance_min = 3, drop_chance_max = 7, setting_id = "stains_slippery_hands" },
  { name = "MATRAN_GOLDRUSH", drop_chance_min = 3, drop_chance_max = 7, setting_id = "stains_goldrush" },
  { name = "MATRAN_PROJECTILE_REPULSION", drop_chance_min = 3, drop_chance_max = 7, setting_id = "stains_projectile_repulsion" },
  { name = "MATRAN_DEFENSE_UP", drop_chance_min = 7, drop_chance_max = 10, setting_id = "stains_defense_up" },
  { name = "MATRAN_ATTRACTIVE", drop_chance_min = 6, drop_chance_max = 8, setting_id = "stains_attractive" },
  { name = "MATRAN_FLOATY", drop_chance_min = 5, drop_chance_max = 7, setting_id = "stains_floaty" },
  { name = "MATRAN_RADIOACTIVE", drop_chance_min = 2, drop_chance_max = 4, setting_id = "stains_radioactive_matran" },
  { name = "MATRAN_BLACKHOLE", drop_chance_min = 3, drop_chance_max = 6, setting_id = "stains_blackhole" },
  { name = "MATRAN_DEMOLITION", drop_chance_min = 4, drop_chance_max = 7, setting_id = "stains_demolition" },
}

-- Gather status effects from other mods and add them to the staining effects table with a random stain drop chance
local function get_active_mods()
  local active_mods = ModGetActiveModIDs()
  for i=#active_mods,1,-1 do
    if active_mods[i] == "material_randomizer" then
      table.remove(active_mods, i)
    end
  end
  return active_mods
end

if get_setting("gather_mod_effects") then
  status_effects = {}
  for i, mod in ipairs(get_active_mods()) do
    local init_content = ModTextFileGetContent("mods/" .. mod .. "/init.lua")
    if init_content then
      local result = init_content:gfind("ModLuaFileAppend%(%s*[\"\']data/scripts/status_effects/status_list.lua[\"\'],%s*[\"\']([^\"\']*)")
      for match, _ in result do
        dofile(match)
      end
    end
  end
  for i, v in ipairs(status_effects) do
    table.insert(stains_effects, { name = v.id, drop_chance_min = 1, drop_chance_max = 5 })
  end
end

for i, effect in ipairs(vanilla_stains_effects) do
  if get_setting(effect.setting_id) then
    table.insert(stains_effects, effect)
  end
end
for i, effect in ipairs(custom_stains_effects) do
  if get_setting(effect.setting_id) then
    table.insert(stains_effects, effect)
  end
end

shuffle(stains_effects)
shuffle(ingestion_effects)

local function remove_stain_effects(element)
  element._attr.status_effects = nil
  element.StatusEffects = element.StatusEffects or {}
  element.StatusEffects.Stains = nil
end

local function remove_ingestion_effects(element)
  element.StatusEffects = element.StatusEffects or {}
  element.StatusEffects.Ingestion = nil
end

local elements_with_effects = {}

local function set_stain_effect(element, stain_effect_id)
  element.StatusEffects.Stains = { StatusEffect = { _attr = {
    type = stain_effect_id,
  } } }
  if get_setting("show_stain_effect_name") then
    elements_with_effects[element] = elements_with_effects[element] or {
       stains = {},
       ingestion = {}
    }
    table.insert(elements_with_effects[element].stains, stain_effect_id)
  end
end

local function set_ingestion_effect(element, ingestion_effect_id, amount)
  element.StatusEffects.Ingestion = { StatusEffect = { _attr = {
    type = ingestion_effect_id,
    amount = amount,
  } } }
  if get_setting("show_ingestion_effect_name") then
    elements_with_effects[element] = elements_with_effects[element] or {
       stains = {},
       ingestion = {}
    }
    table.insert(elements_with_effects[element].ingestion, ingestion_effect_id)
  end
end

-- Give all liquids some random stains + ingestion effects
for i, element in ipairs(categories.liquid) do
  if get_setting("randomize_stain_liquid") then
    local stain_effect = stains_effects[(i % #stains_effects) + 1]
    remove_stain_effects(element)
    if stain_effect then
      element._attr.danger_radioactive = stain_effect == "RADIOACTIVE" and "1" or nil
      element._attr.danger_fire = stain_effect == "ON_FIRE" and "1" or nil
      element._attr.danger_water = stain_effect == "WET" and "1" or nil
      element._attr.danger_poison = stain_effect == "POISONED" and "1" or nil
      local stain_effect_name
      if type(stain_effect) == "table" then
        stain_effect_name = stain_effect.name
        element._attr.liquid_sprite_stain_shaken_drop_chance = rand(stain_effect.drop_chance_min, stain_effect.drop_chance_max)
      else
        stain_effect_name = stain_effect
      end
      set_stain_effect(element, stain_effect_name)
    end
  end
  if get_setting("randomize_ingestion_liquid") then
    local ingestion_effect = ingestion_effects[(i % #ingestion_effects) + 1]
    remove_ingestion_effects(element)
    if ingestion_effect then
      local ingestion_effect_name, ingestion_effect_amount
      if type(ingestion_effect) == "table" then
        ingestion_effect_name = ingestion_effect.name
        ingestion_effect_amount = randf(ingestion_effect.amount_min, ingestion_effect.amount_max)
      else
        ingestion_effect_name = ingestion_effect
        ingestion_effect_amount = randf(0.2, 1.0)
      end
      set_ingestion_effect(element, ingestion_effect_name, ingestion_effect_amount)
    end
  end
end

shuffle(stains_effects)

-- Make it impossible to give gases polymorph effect
local filtered_stains_effects = {}
for i, effect in ipairs(stains_effects) do
  if effect.name ~= "POLYMORPH" and effect.name ~= "POLYMORPH_RANDOM" then
    table.insert(filtered_stains_effects, effect)
  end
end

-- Give some gases random stain effects
for i, element in ipairs(categories.gas) do
  local inherited_element = get_inherited_element(get_tree_node(tree, parents, element._attr.name))
  if get_setting("randomize_stain_gas") and inherited_element._attr.platform_type ~= "1" then
    element._attr.liquid_stains = "1"
    remove_stain_effects(element)
    local stain_effect = filtered_stains_effects[(i % #filtered_stains_effects) + 1]
    if stain_effect then
      element._attr.danger_radioactive = stain_effect == "RADIOACTIVE" and "1" or nil
      element._attr.danger_fire = stain_effect == "ON_FIRE" and "1" or nil
      element._attr.danger_water = stain_effect == "WET" and "1" or nil
      element._attr.danger_poison = stain_effect == "POISONED" and "1" or nil

      local stain_effect_name
      if type(stain_effect) == "table" then
        stain_effect_name = stain_effect.name
        element._attr.liquid_sprite_stain_shaken_drop_chance = rand(stain_effect.drop_chance_min, stain_effect.drop_chance_max)
      else
        stain_effect_name = stain_effect
      end
      set_stain_effect(element, stain_effect_name)
    end
  end
end

shuffle(ingestion_effects)

-- Give some sands random ingestion effects
for i, element in ipairs(categories.sand) do
  if get_setting("randomize_ingestion_sand") then
    remove_ingestion_effects(element)
    if rand(100) <= 25 then
      local ingestion_effect = ingestion_effects[(i % #ingestion_effects) + 1]
      if ingestion_effect then
        local ingestion_effect_name, ingestion_effect_amount
        if type(ingestion_effect) == "table" then
          ingestion_effect_name = ingestion_effect.name
          ingestion_effect_amount = randf(ingestion_effect.amount_min, ingestion_effect.amount_max)
        else
          ingestion_effect_name = ingestion_effect
          ingestion_effect_amount = randf(0.2, 1.0)
        end
        set_ingestion_effect(element, ingestion_effect_name, ingestion_effect_amount)
      end
    end
  end
end

shuffle(ingestion_effects)

-- Give some grounds random ingestion effects
for i, element in ipairs(categories.ground) do
  if get_setting("randomize_ingestion_ground") then
    remove_ingestion_effects(element)
    if rand(100) <= 5 then
      local ingestion_effect = ingestion_effects[(i % #ingestion_effects) + 1]
      if ingestion_effect then
        local ingestion_effect_name, ingestion_effect_amount
        if type(ingestion_effect) == "table" then
          ingestion_effect_name = ingestion_effect.name
          ingestion_effect_amount = randf(ingestion_effect.amount_min, ingestion_effect.amount_max)
        else
          ingestion_effect_name = ingestion_effect
          ingestion_effect_amount = randf(0.2, 1.0)
        end
        set_ingestion_effect(element, ingestion_effect_name, ingestion_effect_amount)
      end
    end
  end
end

shuffle(ingestion_effects)

-- Give fires random ingestion effects
for i, element in ipairs(categories.fire) do
  if get_setting("randomize_ingestion_fire") then
    remove_ingestion_effects(element)
    local ingestion_effect = ingestion_effects[(i % #ingestion_effects) + 1]
    if ingestion_effect then
      local ingestion_effect_name, ingestion_effect_amount
      if type(ingestion_effect) == "table" then
        ingestion_effect_name = ingestion_effect.name
        ingestion_effect_amount = randf(ingestion_effect.amount_min, ingestion_effect.amount_max)
      else
        ingestion_effect_name = ingestion_effect
        ingestion_effect_amount = randf(0.2, 1.0)
      end
      set_ingestion_effect(element, ingestion_effect_name, ingestion_effect_amount)
    end
  end
end

shuffle(ingestion_effects)

-- Give some solids random ingestion effects
for i, element in ipairs(categories.solid) do
  if get_setting("randomize_ingestion_solid") and rand(100) <= 10 then
    remove_ingestion_effects(element)
    local ingestion_effect = ingestion_effects[(i % #ingestion_effects) + 1]
    if ingestion_effect then
      local ingestion_effect_name, ingestion_effect_amount
      if type(ingestion_effect) == "table" then
        ingestion_effect_name = ingestion_effect.name
        ingestion_effect_amount = randf(ingestion_effect.amount_min, ingestion_effect.amount_max)
      else
        ingestion_effect_name = ingestion_effect
        ingestion_effect_amount = randf(0.2, 1.0)
      end
      set_ingestion_effect(element, ingestion_effect_name, ingestion_effect_amount)
    end
  end
end

shuffle(ingestion_effects)

-- Give some breakables random ingestion effects
for i, element in ipairs(categories.breakable) do
  if get_setting("randomize_ingestion_breakable") and rand(100) <= 10 then
    remove_ingestion_effects(element)
    local ingestion_effect = ingestion_effects[(i % #ingestion_effects) + 1]
    if ingestion_effect then
      local ingestion_effect_name, ingestion_effect_amount
      if type(ingestion_effect) == "table" then
        ingestion_effect_name = ingestion_effect.name
        ingestion_effect_amount = randf(ingestion_effect.amount_min, ingestion_effect.amount_max)
      else
        ingestion_effect_name = ingestion_effect
        ingestion_effect_amount = randf(0.2, 1.0)
      end
      set_ingestion_effect(element, ingestion_effect_name, ingestion_effect_amount)
    end
  end
end

local cached_status_names = {}
local function get_status_effect_by_id(id)
  if cached_status_names[id] then
    return cached_status_names[id]
  end
  dofile_once("data/scripts/status_effects/status_list.lua")
  for i, effect in ipairs(status_effects) do
    if effect.id == id then
      cached_status_names[id] = effect
      return effect
    end
  end
end

for element, effects in pairs(elements_with_effects) do
  local stains_string = "Stains: "
  for i, effect_id in ipairs(effects.stains) do
    local effect = get_status_effect_by_id(effect_id)
    stains_string = stains_string .. GameTextGetTranslatedOrNot(effect.ui_name)
    if next(effects.stains, i) then
      stains_string = stains_string .. ", "
    end
  end
  local ingestion_string = "Ingestion: "
  for i, effect_id in ipairs(effects.ingestion) do
    local effect = get_status_effect_by_id(effect_id)
    ingestion_string = ingestion_string .. GameTextGetTranslatedOrNot(effect.ui_name)
    if next(effects.ingestion, i) then
      ingestion_string = ingestion_string .. ", "
    end
  end
  local appendix = ""
  if #effects.stains > 0 then
    appendix = stains_string
  end
  if #effects.ingestion > 0 then
    if #effects.stains > 0 then
      appendix = appendix .. " - " .. ingestion_string
    else
      appendix = ingestion_string
    end
  end
  element._attr.ui_name = GameTextGetTranslatedOrNot(element._attr.ui_name or "") .. (" (%s)"):format(appendix)
end

if get_setting("randomize_reactions") then
  -- Turn two random liquids into the new lava and acid by giving them the tags, which are used by the reaction
  local new_lava_material = random_from_table(categories.liquid)
  local new_acid_material = random_from_table(categories.liquid)
  -- In case we pick the same liquid, try to get a new liquid a maximum of 20 times
  for i=1, 20 do
    if new_acid_material == new_lava_material then
      new_acid_material = random_from_table(categories.liquid)
    end
  end
  add_tag(new_lava_material, "lava")
  print(string.format("%s got the lava reactions.", new_lava_material._attr.name))
  add_tag(new_acid_material, "acid")
  print(string.format("%s got the acid reactions.", new_acid_material._attr.name))

  -- Add custom reactions

  table.insert(handler.root.Materials.Reaction, { _attr = {
    probability="5",
    input_cell1="[shrinking]",
    input_cell2="air",
    output_cell1="air",
    output_cell2="air",
  }})

  table.insert(handler.root.Materials.Reaction, { _attr = {
    probability="20",
    input_cell1="[densify]",
    input_cell2="air",
    output_cell1="matran_super_dense",
    output_cell2="air",
  }})

  local sandify_to_sand = random_from_table(categories.sand)._attr.name
  table.insert(handler.root.Materials.Reaction, { _attr = {
    probability="5",
    input_cell1="[matran_sandify]",
    input_cell2="air",
    output_cell1=sandify_to_sand,
    output_cell2="air",
  }})
  if added_matran_corrosive_gas then
    table.insert(handler.root.Materials.Reaction, { _attr = {
      probability="100",
      input_cell1="[matran_corrosive_gas]",
      input_cell2="[solid]",
      output_cell1="air",
      output_cell2="air",
    }})
    table.insert(handler.root.Materials.Reaction, { _attr = {
      probability="100",
      input_cell1="[matran_corrosive_gas]",
      input_cell2="[static]",
      output_cell1="air",
      output_cell2="air",
    }})
  end

  -- Add a reaction that turns a random sand material into gold
  -- Workaround for not being able to set a certain material to be the new gold.

  if rand(100) <= 5 then
    local gold_sand = random_from_table(categories.sand)._attr.name
    table.insert(handler.root.Materials.ReqReaction, { _attr = {
      probability="100",
      fast_reaction="1",
      input_cell1=gold_sand,
      input_cell2="matran_dummy", -- Use a material that never exists to guarantee this reaction
      output_cell1="gold",
      output_cell2="gold",
    }})
    print(gold_sand .. " turned to gold.")
  end

  local disable_reactions_for_material = {
    "water_swamp", "swamp", "peat"
  }

  for i, reaction in ipairs(handler.root.Materials.Reaction) do
    for i2, material in ipairs(disable_reactions_for_material) do
      if reaction._attr.input_cell1 == material or reaction._attr.input_cell2 == material
         or reaction._attr.output_cell1 == material or reaction._attr.output_cell2 == material then
        table.remove(handler.root.Materials.Reaction, i)
      end
    end
  end
end

local materials_to_add, materials_to_replace
if not EntityCreateNew then
  local t = dofile("new_materials.lua")
  materials_to_add, materials_to_replace = t[1], t[2]
else
  local t = dofile("mods/material_randomizer/files/new_materials.lua")
  materials_to_add, materials_to_replace = t[1], t[2]
end

for i, material in ipairs(materials_to_add) do
  table.insert(handler.root.Materials.CellData, material)
end

-- This isn't local because it gets read out in init.lua to apply it to the player
materials_that_damage = {}

if get_setting("add_new_materials") then
  print("Replacing materials!")
  -- Randomly replace default materials with custom ones

  for category, materials in pairs(materials_to_replace) do
    -- Create #materials non repeating numbers from 1 to #categories[category]
    local all_indexes = {}
    for i=1, #categories[category] do
      table.insert(all_indexes, i)
    end
    local random_indexes = {}
    -- Shuffle the indexes
    for i=1, #materials do
      table.insert(random_indexes, table.remove(all_indexes, rand(1, #all_indexes)))
    end

    for i, replacement_material in ipairs(materials) do
      if rand(100) <= 30 then
        local element = categories[category][random_indexes[i]]
        local old_name = element._attr.name
        local old_wang_color = element._attr.wang_color
        local old_parent = element._attr._parent

        function copy_material(orig, repl)
          for k, v in pairs(orig) do
            orig[k] = nil
          end
          for k, v in pairs(repl) do
            orig[k] = clone_table(v)
          end
        end

        copy_material(element, replacement_material)

        element._attr.name = old_name
        element._attr.wang_color = old_wang_color
        -- Necessary because CellDataChilds needs to have a parent and transforming this into CellData is too much work
        if old_parent then
          element._attr._parent = "matran_dummy"
        end
        -- Filter out custom attributes and apply their effect
        for attr, v in pairs(element._attr) do
          if attr == "contact_damage" then
            materials_that_damage[old_name] = v
            element._attr[attr] = nil
          end
        end
        print("Replaced " .. element._attr.name .. " with " .. replacement_material._attr.name)
      end
    end
  end
end

-- give 2% of liquids random damage
-- materials_that_damage = {
--   acid = 0.005,
--   lava = 0.003,
--   blood_cold_vapour = 0.0006,
--   blood_cold = 0.0009,
--   poison = 0.001,
--   radioactive_gas = 0.001,
--   radioactive_gas_static = 0.001,
--   rock_static_radioactive = 0.001,
--   rock_static_poison = 0.001,
--   ice_radioactive_static = 0.001,
--   ice_radioactive_glass = 0.001,
--   ice_acid_static = 0.001,
--   ice_acid_glass = 0.001,
--   rock_static_cursed = 0.005,
--   --  0.0001 = a little
--   --  0.005 =  a lot
-- }
for i, material in pairs(categories.liquid) do
  if rand(100) <= 2 then
    materials_that_damage[material._attr.name] = rand(1, 10) / 100 / 25
  end
end

if not EntityCreateNew then
  local f = assert(io.open(output_file, "w"))
  f:write(xml2lua.toXml(handler.root, "Materials", 0))
  f:close()
else
  ModTextFileSetContent("data/materials.xml", xml2lua.toXml(handler.root, "Materials", 0))
end
print("Material Randomizer Seed: " .. seed)
print("============== Material Randomizer script end ==============")
