dofile_once("data/scripts/lib/utilities.lua")
dofile_once("data/scripts/lib/mod_settings.lua")

local mod_id = "material_randomizer"
mod_settings_version = 1

local function are_mod_files_available()
	-- Check if we are ingame or at the start menu settings, if we're ingame now, re-check
	if GameGetFrameNum() > 0 and not are_mod_files_available_cached_checked_ingame then
		are_mod_files_available_cached = nil
		are_mod_files_available_cached_checked_ingame = true
	end
	if are_mod_files_available_cached == nil then
		local func = loadfile("mods/" .. mod_id .. "/settings_test.lua")
		are_mod_files_available_cached = func ~= nil
	end
	return are_mod_files_available_cached
end

local function get_setting_id(name)
	return mod_id .. "." .. name
end

local function table_contains(t, value)
	for i, v in ipairs(t) do
		if v == value then
			return true
		end
	end
	return false
end

---@param gui any
---@param id number
---@param x number
---@param y number
---@param checked boolean
---@param disabled boolean
local function gui_checkbox(gui, id, x, y, checked, disabled)
	local checkbox
	if disabled then
		checkbox = "mods/material_randomizer/files/settings/checkbox_invisible.png"
	else
		checkbox = "mods/material_randomizer/files/settings/checkbox" .. (checked and "_fill" or "") .. ".png"
	end
	GuiOptionsAddForNextWidget(gui, GUI_OPTION.NoPositionTween)
	GuiOptionsAddForNextWidget(gui, GUI_OPTION.DrawActiveWidgetCursorOff)
	local clicked, right_clicked = GuiImageButton(gui, id, x, y, "", checkbox)
	return clicked, right_clicked
end

local settings = {
	{
		id = "change_gravity",
		label = "Change gravity",
		description = "Changes the gravity for physics based materials (chunk of ice etc)\nor physics objects made out of those materials",
		value_default = true,
		scope = MOD_SETTING_SCOPE_NEW_GAME,
	},
	{
		id = "gather_mod_effects",
		label = "Include stain & ingestion effects from other mods",
		description = "Does not always work and can cause errors, depends on the mod, try it if you want",
		value_default = true,
		scope = MOD_SETTING_SCOPE_NEW_GAME,
	},
	{
		id = "show_stain_effect_name",
		label = "Show stain effect name",
		description = "Adds the name of the stain effect behind the material name in parantheses,\ne.g.: Parthoonium (Stains: Polymporh - Ingestion: Exploding farts)",
		value_default = false,
		scope = MOD_SETTING_SCOPE_NEW_GAME,
	},
	{
		id = "show_ingestion_effect_name",
		label = "Show ingestion effect name",
		description = "Adds the name of the ingestion effect behind the material name in parantheses,\ne.g.: Parthoonium (Stains: Polymporh - Ingestion: Exploding farts)",
		value_default = false,
		scope = MOD_SETTING_SCOPE_NEW_GAME,
	},
	{
		id = "add_new_materials",
		label = "Add new rare materials",
		description = "Adds new custom materials like milk, uranium, cheese... the cheese one looks really good :)",
		value_default = true,
		scope = MOD_SETTING_SCOPE_NEW_GAME,
	},
	{
		id = "flammable_liquids",
		label = "Flammable liquids",
		description = "Has a low chance to make a liquid flammable",
		value_default = true,
		scope = MOD_SETTING_SCOPE_NEW_GAME,
	},
	{
		id = "randomize_material_damages",
		label = "Randomize material damages",
		description = "Damage like lava, acid etc",
		value_default = true,
		scope = MOD_SETTING_SCOPE_NEW_GAME,
	},
	{
		id = "randomize_reactions",
		label = "Randomize reactions",
		value_default = true,
		scope = MOD_SETTING_SCOPE_NEW_GAME,
	},
	{
		id = "fixed_seed",
		label = "Fixed seed",
		description = "Play with a fixed seed (not world seed, randomizer seed, so you get the same materials etc every time)",
		value_default = false,
		scope = MOD_SETTING_SCOPE_NEW_GAME,
	},
	{ -- Display current seed when a run is in progress
		requires_setting = "fixed_seed",
		ui_fn = function(gui, setting, get_id, in_main_menu)
			if not in_main_menu then
				local seed = ModSettingGet(get_setting_id("current_seed")) or "0" -- Gets set by the randomizer
				local hex_string = string.format("%.8X", seed)
				GuiText(gui, 0, 0, "Current seed: " .. tostring(hex_string))
			end
		end,
	},
	{
		type = "input",
		id = "seed",
		label = "Seed",
		value_default = "0",
		requires_setting = "fixed_seed",
		scope = MOD_SETTING_SCOPE_NEW_GAME,
	},
	{
		type = "title",
		label = "Stain effects",
		description = "Custom status effects you can get from being stained with materials",
	},
	{
		type = "three_columns",
		columns = {
			{
				label = "Positive",
				items = {
					{ id = "stains_infinite_levitation", icon = "mods/material_randomizer/files/effects/infinite_levitation/icon.png", label = "Infinite levitation", value_default = true, color = "green", scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "stains_restore_spell_uses", icon = "mods/material_randomizer/files/effects/restore_spell_uses/icon.png", label = "Restore spell uses", value_default = true, color = "green", scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "stains_rapidfire", icon = "mods/material_randomizer/files/effects/rapidfire/icon.png", label = "Rapidfire", value_default = true, color = "green", scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "stains_attract_projectiles", icon = "mods/material_randomizer/files/effects/attract_projectiles/icon.png", label = "Attract projectiles", value_default = true, color = "green", scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "stains_goldrush", icon = "mods/material_randomizer/files/effects/goldrush/icon.png", label = "Gold rush", value_default = true, color = "green", scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "stains_projectile_repulsion", icon = "mods/material_randomizer/files/effects/projectile_repulsion/icon.png", label = "Projectile repulsion", value_default = true, color = "green", scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "stains_defense_up", icon = "mods/material_randomizer/files/effects/defense_up/icon.png", label = "Defense up", value_default = true, color = "green", scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "stains_attractive", icon = "mods/material_randomizer/files/effects/attractive/icon.png", label = "Attractive", value_default = true, color = "green", scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "stains_floaty", icon = "mods/material_randomizer/files/effects/floaty/icon.png", label = "Floaty", value_default = true, color = "green", scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "stains_radioactive_matran", icon = "mods/material_randomizer/files/effects/radioactive/icon.png", label = "Radioactive", value_default = true, color = "green", scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "stains_blackhole", icon = "mods/material_randomizer/files/effects/blackhole/icon.png", label = "Blackhole", value_default = true, color = "green", scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "stains_demolition", icon = "mods/material_randomizer/files/effects/demolition/icon.png", label = "Demolition", value_default = true, color = "green", scope = MOD_SETTING_SCOPE_NEW_GAME, },
				},
			},
			{
				label = "Negative",
				items = {
					{ id = "stains_slippery_hands", icon = "mods/material_randomizer/files/effects/slippery_hands/icon.png", label = "Slippery hands", value_default = true, color = "red", scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "stains_blindness", icon = "mods/material_randomizer/files/effects/blindness/icon.png", label = "Blindness", value_default = true, color = "red", scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "stains_impotence", icon = "mods/material_randomizer/files/effects/impotence/icon.png", label = "Impotence", value_default = true, color = "red", scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "stains_lose_levitation", icon = "mods/material_randomizer/files/effects/lose_levitation/icon.png", label = "Lose levitation", value_default = true, color = "red", scope = MOD_SETTING_SCOPE_NEW_GAME, },
				},
			},
			{
				label = "Default",
				items = {
					{ id = "stains_wet", icon = "data/ui_gfx/status_indicators/wet.png", label = GameTextGetTranslatedOrNot("$status_wet"), value_default = true, scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "stains_oiled", icon = "data/ui_gfx/status_indicators/oiled.png", label = GameTextGetTranslatedOrNot("$status_oiled"), value_default = true, scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "stains_bloody", icon = "data/ui_gfx/status_indicators/bloody.png", label = GameTextGetTranslatedOrNot("$status_bloody"), value_default = true, scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "stains_slimy", icon = "data/ui_gfx/status_indicators/slimy.png", label = GameTextGetTranslatedOrNot("$status_slimy"), value_default = true, scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "stains_radioactive", icon = "data/ui_gfx/status_indicators/radioactive.png", label = GameTextGetTranslatedOrNot("$status_radioactive"), value_default = true, scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "stains_alcoholic", icon = "data/ui_gfx/status_indicators/alcoholic.png", label = GameTextGetTranslatedOrNot("$status_alcoholic"), value_default = true, scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "stains_poisoned", icon = "data/ui_gfx/status_indicators/poisoned.png", label = GameTextGetTranslatedOrNot("$status_poisoned"), value_default = true, scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "stains_teleportation", icon = "data/ui_gfx/status_indicators/teleportation.png", label = GameTextGetTranslatedOrNot("$status_teleportation"), value_default = true, scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "stains_hp_regeneration", icon = "data/ui_gfx/status_indicators/hp_regeneration.png", label = GameTextGetTranslatedOrNot("$status_hp_regeneration"), value_default = true, scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "stains_polymorph", icon = "data/ui_gfx/status_indicators/polymorph.png", label = GameTextGetTranslatedOrNot("$status_polymorph"), value_default = true, scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "stains_polymorph_random", icon = "data/ui_gfx/status_indicators/polymorph_random.png", label = GameTextGetTranslatedOrNot("$status_polymorph_random"), value_default = true, scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "stains_berserk", icon = "data/ui_gfx/status_indicators/berserk.png", label = GameTextGetTranslatedOrNot("$status_berserk"), value_default = true, scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "stains_charm", icon = "data/ui_gfx/status_indicators/charm.png", label = GameTextGetTranslatedOrNot("$status_charm"), value_default = true, scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "stains_invisibility", icon = "data/ui_gfx/status_indicators/invisibility.png", label = GameTextGetTranslatedOrNot("$status_invisibility"), value_default = true, scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "stains_confusion", icon = "data/ui_gfx/status_indicators/confusion.png", label = GameTextGetTranslatedOrNot("$status_confusion"), value_default = true, scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "stains_movement_faster_2x", icon = "data/ui_gfx/status_indicators/movement_faster.png", label = GameTextGetTranslatedOrNot("$status_movement_faster"), value_default = true, scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "stains_worm_attractor", icon = "data/ui_gfx/status_indicators/worm_attractor.png", label = GameTextGetTranslatedOrNot("$status_worm_attractor"), value_default = true, scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "stains_protection_all", icon = "data/ui_gfx/status_indicators/protection_all.png", label = GameTextGetTranslatedOrNot("$status_protection_all"), value_default = true, scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "stains_mana_regeneration", icon = "data/ui_gfx/status_indicators/mana_regeneration.png", label = GameTextGetTranslatedOrNot("$status_mana_regeneration"), value_default = true, scope = MOD_SETTING_SCOPE_NEW_GAME, },
				},
			},
		}
	},
	{
		type = "title",
		label = "Ingestion effects",
		description = "Custom ingestion effects you can get from eating/drinking materials",
	},
	{
		type = "three_columns",
		columns = {
			{
				label = "Positive",
				items = {
					{ id = "ingestion_minor_regeneration", icon = "mods/material_randomizer/files/effects/minor_regeneration/icon.png", label = "Minor regeneration", value_default = true, color = "green", scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "ingestion_heavy", icon = "mods/material_randomizer/files/effects/heavy/icon.png", label = "Heavy", value_default = true, color = "green", scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "ingestion_powerful_kicks", icon = "mods/material_randomizer/files/effects/powerful_kicks/icon.png", label = "Powerful kicks", value_default = true, color = "green", scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "ingestion_low_gravity", icon = "mods/material_randomizer/files/effects/low_gravity/icon.png", label = "Low gravity", value_default = true, color = "green", scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "ingestion_gain_max_hp", icon = "mods/material_randomizer/files/effects/gain_max_hp/icon.png", label = "Gain max hp", value_default = true, color = "green", scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "ingestion_halve_mana_cost", icon = "mods/material_randomizer/files/effects/halve_mana_cost/icon.png", label = "Halve mana cost", value_default = true, color = "green", scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "ingestion_long_arm", icon = "mods/material_randomizer/files/effects/long_arm/icon.png", label = "Long arm", value_default = true, color = "green", scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "ingestion_anomaly_catalyst", icon = "mods/material_randomizer/files/effects/anomaly_catalyst/icon.png", label = "Anomaly catalyst", value_default = true, color = "green", scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "ingestion_noodle_arm", icon = "mods/material_randomizer/files/effects/noodle_arm/icon.png", label = "Noodle arm", value_default = true, color = "green", scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "ingestion_evaporate_liquids", icon = "mods/material_randomizer/files/effects/evaporate_liquids/icon.png", label = "Evaporate liquids", value_default = true, color = "green", scope = MOD_SETTING_SCOPE_NEW_GAME, },
				},
			},
			{
				label = "Negative",
				items = {
					{ id = "ingestion_incontinence", icon = "mods/material_randomizer/files/effects/incontinence/icon.png", label = "Incontinence", value_default = true, color = "red", scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "ingestion_timebomb", icon = "mods/material_randomizer/files/effects/timebomb/icon.png", label = "Timebomb", value_default = true, color = "red", scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "ingestion_explosive_farts", icon = "mods/material_randomizer/files/effects/explosive_farts/icon.png", label = "Explosive farts", value_default = true, color = "red", scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "ingestion_diarrhea", icon = "mods/material_randomizer/files/effects/diarrhea/icon.png", label = "Diarrhea", value_default = true, color = "red", scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "ingestion_bleed_gold", icon = "mods/material_randomizer/files/effects/bleed_gold/icon.png", label = "Bleed gold", value_default = true, color = "red", scope = MOD_SETTING_SCOPE_NEW_GAME, },
				},
			},
			{
				label = "Default",
				items = {
					{ id = "ingestion_drunk", icon = "data/ui_gfx/status_indicators/drunk.png", label = "Drunk", value_default = true, scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "ingestion_trip", icon = "data/ui_gfx/status_indicators/trip.png", label = "Tripping", value_default = true, scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "ingestion_nightvision", icon = "data/ui_gfx/status_indicators/nightvision.png", label = "Wormy vision", value_default = true, scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "ingestion_food_poisoning", icon = "data/ui_gfx/status_indicators/food_poisoning.png", label = "Food poisoning", value_default = false, scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "ingestion_movement_slower", icon = "data/ui_gfx/status_indicators/ingestion_movement_slower.png", label = "Clumsy movement", value_default = false, scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "ingestion_freezing", icon = "data/ui_gfx/status_indicators/ingestion_freezing.png", label = "Chilly", value_default = false, scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "ingestion_farts", icon = "data/ui_gfx/status_indicators/farts.png", label = "Farts", value_default = false, scope = MOD_SETTING_SCOPE_NEW_GAME, },
					{ id = "ingestion_on_fire", icon = "data/ui_gfx/status_indicators/internal_fire.png", label = "On fire", value_default = false, scope = MOD_SETTING_SCOPE_NEW_GAME, },
				},
			},
		},
	},
	{
		type = "title",
		label = "Randomization Matrix",
		description = "Here you can set which things to randomize",
	},
	{ -- the setting id will be made up of the setting id, row id and column id like: randomize_name_ground
		id = "randomize",
		type = "matrix",
		value_default = {
			stain_gas = false,
		},
		scope = MOD_SETTING_SCOPE_NEW_GAME,
		columns = {
			{ id = "ground", label = "Ground", disabled_with = { "stain" } },
      { id = "liquid", label = "Liquid", },
      { id = "sand", label = "Sand", disabled_with = { "stain" } },
      { id = "gas", label = "Gas", disabled_with = { "ingestion" } },
      { id = "fire", label = "Fire", disabled_with = { "stain" } },
      { id = "solid", label = "Solid (Physics objects, glass etc)", disabled_with = { "stain" } },
      { id = "breakable", label = "Breakable", disabled_with = { "stain" } },
		},
		rows = {
			{ id = "properties", label = "Properties" },
			{ id = "name", label = "Names" },
			{ id = "visual", label = "Visuals" },
			{ id = "stain", label = "Stain effects" },
			{ id = "ingestion", label = "Ingestion effects" },
		},
	},
	{
		type = "title",
		label = "Exclude materials from randomization",
	},
	{ id = "exclude_material_blood", label = "Blood", value_default = false, scope = MOD_SETTING_SCOPE_NEW_GAME, },
	{ id = "exclude_material_water", label = "Water", value_default = false, scope = MOD_SETTING_SCOPE_NEW_GAME, },
	{ id = "exclude_material_radioactive_liquid", label = "Toxic sludge", value_default = false, scope = MOD_SETTING_SCOPE_NEW_GAME, },
	{
		ui_fn = function(gui, setting, get_id)
			local function string_starts_with(String,Start)
				return string.sub(String,1,string.len(Start))==Start
			end
			
			if GuiButton(gui, get_id(), 0, 0, "Revert settings to defaults") then
				local count = ModSettingGetCount()
				local this_setting_ids = {}
				for i=0, count-1 do
					local setting_id = ModSettingGetAtIndex(i)
					if string_starts_with(setting_id, mod_id) then
						table.insert(this_setting_ids, setting_id)
					end
				end
				for i, setting_id in ipairs(this_setting_ids) do
					ModSettingRemove(setting_id)
				end
			end
		end
	},
}

function ModSettingsGuiCount()
	return 1
end

function ModSettingsUpdate(init_scope)
	local function set_defaults(setting)
		if setting.type == "group" then
			for i, item in ipairs(setting.items) do
				set_defaults(item)
			end
		elseif setting.type == "three_columns" then
			for i, column in ipairs(setting.columns) do
				for i, setting in ipairs(column.items) do
					local setting_id = get_setting_id(setting.id)
					ModSettingSetNextValue(setting_id, setting.value_default, true)
				end
			end
		elseif setting.type == "matrix" then
			for key, default in pairs(setting.value_default) do
				local setting_id = get_setting_id(setting.id .. "_" .. key)
				ModSettingSetNextValue(setting_id, default, true)
			end
			for column in ipairs(setting.columns) do
				for row in ipairs(setting.rows) do
					local setting_id = get_setting_id(setting.id .. "_" .. setting.rows[row].id .. "_" .. setting.columns[column].id)
					ModSettingSetNextValue(setting_id, true, true)
				end
			end
		else
			if setting.value_default ~= nil then
				ModSettingSetNextValue(get_setting_id(setting.id), setting.value_default, true)
			end
		end
	end

	local function save_setting(setting)
		if setting.type == "group" then
			for i, item in ipairs(setting.items) do
				save_setting(item)
			end
		elseif setting.type == "three_columns" then
			for i, column in ipairs(setting.columns) do
				for i, setting in ipairs(column.items) do
					if setting.scope ~= nil and setting.scope >= init_scope then
						local setting_id = get_setting_id(setting.id)
						local next_value = ModSettingGetNextValue(setting_id)
						ModSettingSet(setting_id, next_value)
					end
				end
			end
		elseif setting.type == "matrix" then
			for column in ipairs(setting.columns) do
				for row in ipairs(setting.rows) do
					local setting_id = get_setting_id(setting.id .. "_" .. setting.rows[row].id .. "_" .. setting.columns[column].id)
					local next_value = ModSettingGetNextValue(setting_id)
					ModSettingSet(setting_id, next_value)
				end
			end
		elseif setting.id ~= nil and setting.scope ~= nil and setting.scope >= init_scope then
			local next_value = ModSettingGetNextValue(get_setting_id(setting.id))
			if next_value ~= nil then
				ModSettingSet(get_setting_id(setting.id), next_value)
			end
		end
	end
	for i, setting in ipairs(settings) do
		set_defaults(setting)
		save_setting(setting)
	end
	ModSettingSet(get_setting_id("_version"), mod_settings_version)
end

function ModSettingsGui( gui, in_main_menu )
	local id = 0
	local function get_id()
		id = id + 1
		return id
	end

	if not are_mod_files_available() then
		GuiText(gui, 0, 0, "These mod settings need access to the mod files,\nwhich are not loaded in the main menu.\nPlease start a new game or continue a run\nto access these mod settings.")
		return
	end

	for i, setting in ipairs(settings) do
		if not setting.requires_setting or ModSettingGetNextValue(get_setting_id(setting.requires_setting)) then
			if setting.ui_fn then
				setting.ui_fn(gui, setting, get_id, in_main_menu)
			elseif setting.type == "group" then
				GuiOptionsAddForNextWidget(gui, GUI_OPTION.DrawSemiTransparent)
				GuiText(gui, 0, 0, setting.label)
				if setting.description then
					GuiTooltip(gui, setting.description, "")
				end
				GuiLayoutBeginHorizontal(gui, 0, 0)
				local offset = setting.offset or 0
	
				-- Render labels
				GuiLayoutBeginVertical(gui, 0, 0)
				for i, setting in ipairs(setting.items) do
					GuiText(gui, 0, 0, setting.label)
				end
				GuiLayoutEnd(gui)
	
				GuiLayoutBeginVertical(gui, 0, 0)
				-- Render sliders
				for i, setting in ipairs(setting.items) do
					local next_value = ModSettingGetNextValue(get_setting_id(setting.id))
					local new_value = GuiSlider(gui, get_id(), 10 + offset, 2, "", next_value, setting.value_min, setting.value_max, setting.value_default, setting.value_display_multiplier, " " or setting.value_display_formatting, 80)
					GuiLayoutAddVerticalSpacing(gui, 1)
					if new_value ~= next_value then
						ModSettingSetNextValue(get_setting_id(setting.id), new_value, false)
					end
				end
				GuiLayoutEnd(gui)
	
				-- Render values
				GuiLayoutBeginVertical(gui, 0, 0)
				for i, setting in ipairs(setting.items) do
					if setting.value_min and setting.value_max then
						local next_value = ModSettingGetNextValue(get_setting_id(setting.id))
						GuiText(gui, offset + 30, 0, setting.format_fn and setting.format_fn(next_value) or tostring(next_value))
					else
						GuiText(gui, 0, 0, "")
					end
				end
				GuiLayoutEnd(gui)
	
				GuiLayoutEnd(gui)
				-- Need to do this because the game doesn't count how many items are in the vertical group
				for i=2, #setting.items do
					GuiText(gui, 0, 0, " ")
				end
				-- A little margin at the bottom before the next group or items
				GuiLayoutAddVerticalSpacing(gui, 5)
			elseif setting.type == "input" then
				local setting_id = get_setting_id(setting.id)
				local current_value = ModSettingGetNextValue(setting_id) or ""
				GuiLayoutBeginHorizontal(gui, 0, 0)
				GuiText(gui, 0, 0, setting.label)
				local new_value = GuiTextInput(gui, get_id(), 2, 0, current_value, 55, 8, "0123456789abcdefABCDEF")
				GuiLayoutEnd(gui)
				if current_value ~= new_value then
					ModSettingSetNextValue(setting_id, new_value:upper(), false)
				end
			elseif setting.type == "title" then
				GuiOptionsAddForNextWidget(gui, GUI_OPTION.DrawSemiTransparent)
				GuiText(gui, 0, 0, setting.label)
				if setting.description then
					GuiTooltip(gui, setting.description, "")
				end
			elseif setting.type == "three_columns" then -- Too lazy to generalize it and make it work with all types of widgets
				local column_widths = { 0, 0, 0 }
				local highest_item_count_vertical = 0
				GuiLayoutBeginHorizontal(gui, 0, 0)
				for i, column in ipairs(setting.columns) do
					highest_item_count_vertical = math.max(highest_item_count_vertical, #column.items)
					local column_x = 0
					if i == 2 then
						column_x = column_widths[1]
					elseif i == 3 then
						column_x = column_widths[1] + column_widths[2]
					end
					GuiLayoutBeginVertical(gui, column_x, 0, true) -- Column
					GuiText(gui, 0, 0, column.label)
					for _, setting in ipairs(column.items) do
						local text_width = GuiGetTextDimensions(gui, setting.label)
						column_widths[i] = math.max(column_widths[i], text_width)
						local setting_id = get_setting_id(setting.id)
						local current_value = ModSettingGetNextValue(setting_id)
						local checkbox_clicked, checkbox_right_clicked = gui_checkbox(gui, get_id(), 0, 1, current_value)
						local _, _, _, x, y = GuiGetPreviousWidgetInfo(gui)
						if setting.icon then
							GuiOptionsAddForNextWidget(gui, GUI_OPTION.Layout_NoLayouting)
							GuiImage(gui, get_id(), x + 9, y - 2, setting.icon, 1, 1, 1)
							x = x + 10
						end
						GuiOptionsAddForNextWidget(gui, GUI_OPTION.Layout_NoLayouting)
						local label_clicked, label_right_clicked = GuiButton(gui, get_id(), x + 12, y - 1, setting.label)
						if setting.description then
							GuiTooltip(gui, setting.description, "")
						end
						if checkbox_clicked or label_clicked then
							ModSettingSetNextValue(setting_id, not current_value, false)
						end
						GuiLayoutAddVerticalSpacing(gui, 1)
					end
					GuiLayoutEnd(gui) -- Column
				end
				GuiLayoutEnd(gui)
				for i=1, highest_item_count_vertical do
					GuiText(gui, 0, 0, " ")
				end
			elseif type(setting.value_default) == "boolean" then -- Render Checkbox + Label
				local setting_id = get_setting_id(setting.id)
				local current_value = ModSettingGetNextValue(setting_id)
				GuiLayoutBeginHorizontal(gui, 0, 0)
				local checkbox_clicked, checkbox_right_clicked = gui_checkbox(gui, get_id(), 0, 1, current_value)
				if setting.color == "green" then
					GuiColorSetForNextWidget(gui, 0.25, 0.8, 0.25, 1.0)
				elseif setting.color == "red" then
					GuiColorSetForNextWidget(gui, 0.8, 0.25, 0.25, 1.0)
				end
				local label_clicked, label_right_clicked = GuiButton(gui, get_id(), 1, 0, setting.label)
				if setting.description then
					GuiTooltip(gui, setting.description, "")
				end
				if checkbox_clicked or label_clicked then
					ModSettingSetNextValue(setting_id, not current_value, false)
				end
				GuiLayoutEnd(gui)
				GuiLayoutAddVerticalSpacing(gui, -2)
			elseif setting.type == "matrix" then
				GuiLayoutAddVerticalSpacing(gui, 24)
				local _, _, _, topleft_x, topleft_y, _, text_height = GuiGetPreviousWidgetInfo(gui)
				topleft_y = topleft_y + text_height + 4
				GuiLayoutBeginVertical(gui, 0, 0)
				local widest_text = 0
				for row=1, #setting.rows do
					GuiText(gui, 0, 0, setting.rows[row].label)
					widest_text = math.max(widest_text, GuiGetTextDimensions(gui, setting.rows[row].label))
				end
				GuiLayoutEnd(gui)
				for column=1, #setting.columns do
					GuiOptionsAddForNextWidget(gui, GUI_OPTION.Layout_NoLayouting)
					local spacing = 2
					GuiImage(gui, get_id(), topleft_x + spacing + widest_text + (column-1) * (16 + spacing), topleft_y, "mods/material_randomizer/files/settings/"..setting.columns[column].id..".png", 1, 1, 1)
					GuiTooltip(gui, setting.columns[column].label, "")
					for row=1, #setting.rows do
						local setting_id = get_setting_id(setting.id .. "_" .. setting.rows[row].id .. "_" .. setting.columns[column].id)
						local current_value = not not ModSettingGetNextValue(setting_id)
						local checked = current_value
						local disabled = table_contains(setting.columns[column].disabled_with or {}, setting.rows[row].id)
						GuiOptionsAddForNextWidget(gui, GUI_OPTION.Layout_NoLayouting)
						local clicked, right_clicked = gui_checkbox(gui, get_id(), topleft_x + spacing + 4.5 + widest_text + (column-1) * (16 + spacing), topleft_y + 10 + row * text_height, checked, disabled)
						if clicked then
							ModSettingSetNextValue(setting_id, not current_value, false)
						end
					end
				end
				-- Need to do this because the game doesn't count how many items are in the vertical group
				for row=1, #setting.rows-1 do
					GuiText(gui, 0, 0, " ")
				end
			else
				if type(setting.value_default) == "number" then
					local next_value = ModSettingGetNextValue(get_setting_id(setting.id))
					local new_value = GuiSlider(gui, get_id(), 0, 0, setting.label, next_value, setting.value_min, setting.value_max, setting.value_default, setting.value_display_multiplier, setting.value_display_formatting, 80)
					if new_value ~= next_value then
						ModSettingSetNextValue(get_setting_id(setting.id), new_value, false)
					end
				end
			end
		end
	end
end
